#!/usr/bin/env python
#python 3
#Contact, Adam Taranto, aptaranto@ucdavis.edu

##########################################################################
# Find soft-clipped alignments containing unassembled telomeric repeats. #
##########################################################################

import os
import sys
import re

__version__ = "0.0.2"

class Error (Exception): pass

def isfile(path):
    """
    Test for existence of input file.
    """
    path = os.path.abspath(path)
    if not os.path.isfile(path):
        print("Input file not found: %s" % path)
        sys.exit(1)
    else:
        return path

def read_fai(fai):
    """
    Import fasta index file. Return dict of sequence names and lengths.
    """
    path = isfile(fai)
    # Init empty dict
    ContigDict = dict()
    # Read fai_file to dict
    with open(path, "rU") as f:
        for line in f.readlines():
            li = line.strip().split()
            ContigDict[li[0]]=int(li[1])
    return ContigDict

def addRevComplement(motifList):
    """
    Take list of DNA motif strings and return unique set of strings and their reverse complements.
    """
    revcompl = lambda x: ''.join([{'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N':'N'}[B] for B in x][::-1])
    setList = list()
    for motif in motifList:
        setList.append(motif)
        setList.append(revcompl(motif))
    return set(setList)

def splitCIGAR(SAM_CIGAR):
    """
    Split CIGAR string into list of tuples with format (len,operator)
    """
    CIGARlist = list()
    for x in re.findall('[0-9]*[A-Z|=]',SAM_CIGAR):
        CIGARlist.append((int(re.findall('[0-9]*', x)[0]),re.findall('[A-Z]|=', x)[0]))
    #174M76S --> [(174,M),(76,S)]
    #96S154M --> [(96,S),(154,M)]
    return CIGARlist

def lenCIGAR(SAM_CIGAR):
    """
    Calculate length of alignment in reference sequence as sum of 
    match, read-deletion, splice, mismatch, and read-match block values.
    Ignore read-insertions, padding, hard and soft clip blocks.
    """
    alnLen = 0
    CIGARlist = splitCIGAR(SAM_CIGAR)
    for x in CIGARlist: # i.e. = [(174,M),(76,S)]
        if x[1] in set(['D','M','N','X','=']):
            alnLen += x[0]
    #Ignore operators in set('P','H','S','I')
    return alnLen

def checkClips(SAM_CIGAR):
    """
    Get lengths of soft-clipped blocks from either end of an alignment given a CIGAR string.
    """
    leftClipLen = None
    rightClipLen = None
    CIGARlist = splitCIGAR(SAM_CIGAR)
    # Check if first segment is soft-clipped
    if CIGARlist[0][1] == "S" :
        leftClipLen = int(CIGARlist[0][0])
    # Check if last segment is soft-clipped
    if CIGARlist[-1][1] == "S" :
        rightClipLen = int(CIGARlist[-1][0])
    return (leftClipLen,rightClipLen)

def isClipMotif(samline,motifList,leftClip,rightClip,leftClipLen,rightClipLen):
    """
    Extract terminal soft-clipped blocks from read sequence and test for presence of any DNA motif in motifList.
    """
    clipSeq = list()
    SAM_SEQ = 9
    if leftClip:
        clipSeq.append(samline[SAM_SEQ][0:leftClipLen])
    if rightClip:
        clipSeq.append(samline[SAM_SEQ][-rightClipLen:])
    # True if either clipped end sequence contains at least one instance of any motif
    return any(s in x for s in motifList for x in clipSeq)

"""
CIGAR Operators
--------------------
D   Deletion; the nucleotide is present in the reference but not in the read
H   Hard Clipping; the clipped nucleotides are not present in the read.
I   Insertion; the nucleotide is present in the read  but not in the rference.
M   Match; can be either an alignment match or mismatch. The nucleotide is present in the reference.
N   Skipped region; a region of nucleotides is not present in the read
P   Padding; padded area in the read and not in the reference
S   Soft Clipping;  the clipped nucleotides are present in the read
X   Read Mismatch; the nucleotide is present in the reference
=   Read Match; the nucleotide is present in the reference

The output order in the array is “MIDNSHP=X” followed by a field for the NM tag. If the NM tag is not present, this field will always be 0.

M   BAM_CMATCH  0
I   BAM_CINS    1
D   BAM_CDEL    2
N   BAM_CREF_SKIP   3
S   BAM_CSOFT_CLIP  4
H   BAM_CHARD_CLIP  5
P   BAM_CPAD    6
=   BAM_CEQUAL  7
X   BAM_CDIFF   8
B   BAM_CBACK   9
NM  NM tag  10  
"""



#!/usr/bin/env python

from __future__ import print_function
from teloclip import __version__
import teloclip
import argparse
import sys

def log(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def mainArgs():
    parser = argparse.ArgumentParser(description='Filter SAM file for clipped alignments containing unassembled telomeric repeats.',prog='teloclip')
    # Input options
    parser.add_argument('samfile', nargs='?', type=argparse.FileType('rU'), default=sys.stdin)
    parser.add_argument('--refIdx',type=str,required=True,help='Path to fai index for reference fasta. Index fasta using `samtools faidx FASTA`')
    parser.add_argument('--minClip',type=int,default=1,help='Require clip to extend past ref contig end by at least N bases.')
    parser.add_argument('--maxBreak',type=int,default=50,help='Tolerate max N unaligned bases at contig ends.')
    parser.add_argument('--motifs',type=str,default=None, help='If set keep only reads containing given motif/s from comma delimited list of strings. \
    By default also search for reverse complement of motifs. i.e. TTAGGG,TTAAGGG will also match CCCTAA,CCCTTAA')
    parser.add_argument('--norev',default=False,action='store_true', help='If set do NOT search for reverse complement of specified motifs.')
    parser.add_argument('--matchAny',default=False,action='store_true', help='If set motif match may occur in unclipped region of alignment.')
    parser.add_argument('--version', action='version',version='%(prog)s {version}'.format(version=__version__))
    args = parser.parse_args()
    return args

def main():
    # Get cmd line args
    args = mainArgs()
    # SAM line index keys
    SAM_QNAME   = 0
    SAM_RNAME   = 2
    SAM_POS     = 3
    SAM_CIGAR   = 5
    SAM_SEQ     = 9
    # Start counters
    bothCount = 0
    keepCount = 0
    motifCount = 0
    removeCount = 0
    samlineCount = 0
    # Fetch contig lens
    ContigDict = teloclip.read_fai(args.refIdx)
    # Split motifs into list
    if args.motifs:
        motifList = args.motifs.split(",")
        # Add rev comp motifs to list, make unique set.
        if not args.norev:
            motifList = teloclip.addRevComplement(motifList)
    # Read SAM from stdin
    for line in args.samfile:
        keepLine = False
        leftClip = False
        rightClip = False
        if line[0][0] == "@":
            sys.stdout.write(line)
            continue
        samlineCount += 1
        samline = line.split('\t')
        # Check if line contains soft-clip and no hard-clipping.
        if "S" in samline[SAM_CIGAR] and not "H" in samline[SAM_CIGAR]:
            leftClipLen,rightClipLen = teloclip.checkClips(samline[SAM_CIGAR])
            alnLen = teloclip.lenCIGAR(samline[SAM_CIGAR])
            # Check for left overhang
            if leftClipLen:
                if (int(samline[SAM_POS]) <= args.maxBreak) and (leftClipLen >= (int(samline[SAM_POS]) + args.minClip)):
                    # Overhang is on contig left
                    keepLine = True
                    leftClip = True
                    keepCount += 1
            # Check for right overhang
            if rightClipLen:
                try: ContigLen = ContigDict[str(samline[SAM_RNAME])]
                except: sys.exit("Reference sequence not found in FAI file: " + str(samline[SAM_RNAME]))
                alnEnd = int(samline[SAM_POS]) + alnLen 
                # Check if overhang is on contig right end
                if ((ContigLen - alnEnd) <= args.maxBreak) and (alnEnd + rightClipLen >= ContigLen +1 ):
                    rightClip = True
                    # Check if already found left OH
                    if not keepLine:
                        keepLine = True
                        keepCount += 1
                    else: 
                        # Print to stderr
                        log(str(samline[SAM_QNAME]) + " overhang on both ends of " + str(samline[SAM_RNAME]))
                        bothCount += 1
            # Optional check for Telomeric repeat motifs
            if args.motifs and keepLine and args.matchAny:
                if any(s in samline[SAM_SEQ] for s in motifList):
                    sys.stdout.write(line)
                    motifCount += 1
                else:
                    removeCount += 1
            elif args.motifs and keepLine:
                if teloclip.isClipMotif(samline,motifList,leftClip,rightClip,leftClipLen,rightClipLen):
                    sys.stdout.write(line)
                    motifCount += 1
                else:
                    removeCount += 1
            elif keepLine:  
                sys.stdout.write(line)
            else:
                removeCount += 1
    if args.motifs:
        log("Processes %s SAM records. \nFound %s alignments soft-clipped at contig ends. \nOutput %s alignments containing motif matches. \nDiscarded %s terminal alignments after filtering." % (samlineCount,keepCount,motifCount,removeCount))
    else:
        log("Processes %s SAM records. \nFound %s alignments soft-clipped at contig ends. \nFound %s alignments spanning entire contigs. \nDiscarded %s terminal alignments after filtering." % (samlineCount,keepCount,bothCount,removeCount))

"""
# SAM format
# ----------------------------------------------------------------
# 0   QNAME   String  # Query template NAME
# 1   FLAG    Int     # Bitwise FLAG
# 2   RNAME   String  # References sequence NAME
# 3   POS     Int     # 1-based leftmost mapping POSition
# 4   MAPQ    Int     # MAPping Quality
# 5   CIGAR   String  # CIGAR String
# 6   RNEXT   String  # Ref. name of the mate/next read
# 7   PNEXT   Int     # Position of the mate/next read
# 8   TLEN    Int     # Observed Template LENgth
# 9   SEQ     String  # Segment SEQuence
# 10  QUAL    String  # ASCII of Phred-scaled base QUALity+33
"""